package jsr.daemon;

import jsr.daemon.logging.Logging;

/**
 * interface for the service providing container in order to configure a daemon
 *
 * @author   Thomas Kopp
 * @version  0.9, 2000-11-16
 */
public interface DaemonContext {
    
    /**
     * method for naming a daemon
     * a given name can be retrieved via the configuration interface
     * @param name the name given to a daemon
     */
    public void setName(String name);

    /**
     * method for supplying a logging interface for a daemon 
     * log calls in the logging interface are mapped to no operation if no logger is set
     * @param logger underlying logging resource to be used by a daemon
     */
    public void setLogger(Logging logger);
    
    /**
     * optional method for initializing the size of a daemon configuration
     * method supplies an estimation for allocating resources
     * @param initialCapacity estimated number of attributes supplied by the context
     */
    public void setCapacity(int initialCapacity);
    
    /**
     * method for setting an attribute of the daemon configuration
     * attributes can be retrieved via the configuration interface
     * @param name the name of the attribute to be set
     * @param value the value of the attribute to be set
     */
    public void setAttribute(String name, Object value);
    
    /**
     * method for removing an attribute from a daemon configuration
     * @param name the name of the attribute to be removed
     */
    public void removeAttribute(String name);
}
